@@ -8,16 +8,22 @@ import time
 import csv
 import decimal
 from decimal import Decimal
-import concurrent.futures
-import traceback
-from typing import Sequence
+from typing import Sequence, Optional
+
+from aiorpcx.curio import timeout_after, TaskTimeout, TaskGroup
 
 from .bitcoin import COIN
 from .i18n import _
-from .util import (PrintError, ThreadJob, make_dir, log_exceptions,
+from .util import (ThreadJob, make_dir, log_exceptions,
                    make_aiohttp_session, resource_path)
 from .network import Network
 from .simple_config import SimpleConfig
+from .logging import Logger
+
+
+DEFAULT_ENABLED = False
+DEFAULT_CURRENCY = "EUR"
+DEFAULT_EXCHANGE = "CoinGecko"  # default exchange should ideally provide historical rates
 
 
 # See https://en.wikipedia.org/wiki/ISO_4217
@@ -29,9 +35,10 @@ CCY_PRECISIONS = {'BHD': 3, 'BIF': 0, 'BYR': 0, 'CLF': 4, 'CLP': 0,
                   'VUV': 0, 'XAF': 0, 'XAU': 4, 'XOF': 0, 'XPF': 0}
 
 
-class ExchangeBase(PrintError):
+class ExchangeBase(Logger):
 
     def __init__(self, on_quotes, on_history):
+        Logger.__init__(self)
         self.history = {}
         self.quotes = {}
         self.on_quotes = on_quotes
@@ -40,15 +47,21 @@ class ExchangeBase(PrintError):
     async def get_raw(self, site, get_string):
         # APIs must have https
         url = ''.join(['https://', site, get_string])
-        async with make_aiohttp_session(Network.get_instance().proxy) as session:
+        network = Network.get_instance()
+        proxy = network.proxy if network else None
+        async with make_aiohttp_session(proxy) as session:
             async with session.get(url) as response:
+                response.raise_for_status()
                 return await response.text()
 
     async def get_json(self, site, get_string):
         # APIs must have https
         url = ''.join(['https://', site, get_string])
-        async with make_aiohttp_session(Network.get_instance().proxy) as session:
+        network = Network.get_instance()
+        proxy = network.proxy if network else None
+        async with make_aiohttp_session(proxy) as session:
             async with session.get(url) as response:
+                response.raise_for_status()
                 # set content_type to None to disable checking MIME type
                 return await response.json(content_type=None)
 
@@ -60,46 +73,41 @@ class ExchangeBase(PrintError):
     def name(self):
         return self.__class__.__name__
 
-    @log_exceptions
     async def update_safe(self, ccy):
         try:
-            self.print_error("getting fx quotes for", ccy)
+            self.logger.info(f"getting fx quotes for {ccy}")
             self.quotes = await self.get_rates(ccy)
-            self.print_error("received fx quotes")
+            self.logger.info("received fx quotes")
         except BaseException as e:
-            self.print_error("failed fx quotes:", repr(e))
+            self.logger.info(f"failed fx quotes: {repr(e)}")
             self.quotes = {}
         self.on_quotes()
 
-    def update(self, ccy):
-        asyncio.get_event_loop().create_task(self.update_safe(ccy))
-
-    def read_historical_rates(self, ccy, cache_dir):
+    def read_historical_rates(self, ccy, cache_dir) -> Optional[dict]:
         filename = os.path.join(cache_dir, self.name() + '_'+ ccy)
-        if os.path.exists(filename):
-            timestamp = os.stat(filename).st_mtime
-            try:
-                with open(filename, 'r', encoding='utf-8') as f:
-                    h = json.loads(f.read())
-                h['timestamp'] = timestamp
-            except:
-                h = None
-        else:
-            h = None
-        if h:
-            self.history[ccy] = h
-            self.on_history()
+        if not os.path.exists(filename):
+            return None
+        timestamp = os.stat(filename).st_mtime
+        try:
+            with open(filename, 'r', encoding='utf-8') as f:
+                h = json.loads(f.read())
+        except:
+            return None
+        if not h:  # e.g. empty dict
+            return None
+        h['timestamp'] = timestamp
+        self.history[ccy] = h
+        self.on_history()
         return h
 
     @log_exceptions
     async def get_historical_rates_safe(self, ccy, cache_dir):
         try:
-            self.print_error("requesting fx history for", ccy)
+            self.logger.info(f"requesting fx history for {ccy}")
             h = await self.request_history(ccy)
-            self.print_error("received fx history for", ccy)
+            self.logger.info(f"received fx history for {ccy}")
         except BaseException as e:
-            self.print_error("failed fx history:", e)
-            #traceback.print_exc()
+            self.logger.info(f"failed fx history: {repr(e)}")
             return
         filename = os.path.join(cache_dir, self.name() + '_' + ccy)
         with open(filename, 'w', encoding='utf-8') as f:
@@ -123,28 +131,26 @@ class ExchangeBase(PrintError):
     def historical_rate(self, ccy, d_t):
         return self.history.get(ccy, {}).get(d_t.strftime('%Y-%m-%d'), 'NaN')
 
-    def get_currencies(self):
-        rates = self.get_rates('')
+    async def request_history(self, ccy):
+        raise NotImplementedError()  # implemented by subclasses
+
+    async def get_rates(self, ccy):
+        raise NotImplementedError()  # implemented by subclasses
+
+    async def get_currencies(self):
+        rates = await self.get_rates('')
         return sorted([str(a) for (a, b) in rates.items() if b is not None and len(a)==3])
 
+
 class BitcoinAverage(ExchangeBase):
+    # note: historical rates used to be freely available
+    # but this is no longer the case. see #5188
 
     async def get_rates(self, ccy):
         json = await self.get_json('apiv2.bitcoinaverage.com', '/indices/global/ticker/short')
         return dict([(r.replace("BTC", ""), Decimal(json[r]['last']))
                      for r in json if r != 'timestamp'])
 
-    def history_ccys(self):
-        return ['AUD', 'BRL', 'CAD', 'CHF', 'CNY', 'EUR', 'GBP', 'IDR', 'ILS',
-                'MXN', 'NOK', 'NZD', 'PLN', 'RON', 'RUB', 'SEK', 'SGD', 'USD',
-                'ZAR']
-
-    async def request_history(self, ccy):
-        history = await self.get_csv('apiv2.bitcoinaverage.com',
-                               "/indices/global/history/BTC%s?period=alltime&format=csv" % ccy)
-        return dict([(h['DateTime'][:10], h['Average'])
-                     for h in history])
-
 
 class Bitcointoyou(ExchangeBase):
 
@@ -152,9 +158,6 @@ class Bitcointoyou(ExchangeBase):
         json = await self.get_json('bitcointoyou.com', "/API/ticker.aspx")
         return {'BRL': Decimal(json['ticker']['last'])}
 
-    def history_ccys(self):
-        return ['BRL']
-
 
 class BitcoinVenezuela(ExchangeBase):
 
@@ -210,9 +213,14 @@ class Bitso(ExchangeBase):
 
 class BitStamp(ExchangeBase):
 
+    async def get_currencies(self):
+        return ['USD', 'EUR']
+
     async def get_rates(self, ccy):
-        json = await self.get_json('www.bitstamp.net', '/api/ticker/')
-        return {'USD': Decimal(json['last'])}
+        if ccy in CURRENCIES[self.name()]:
+            json = await self.get_json('www.bitstamp.net', f'/api/v2/ticker/btc{ccy.lower()}/')
+            return {ccy: Decimal(json['last'])}
+        return {}
 
 
 class Bitvalor(ExchangeBase):
@@ -229,18 +237,11 @@ class BlockchainInfo(ExchangeBase):
         return dict([(r, Decimal(json[r]['15m'])) for r in json])
 
 
-class BTCChina(ExchangeBase):
+class Bylls(ExchangeBase):
 
     async def get_rates(self, ccy):
-        json = await self.get_json('data.btcchina.com', '/data/ticker')
-        return {'CNY': Decimal(json['ticker']['last'])}
-
-
-class BTCParalelo(ExchangeBase):
-
-    async def get_rates(self, ccy):
-        json = await self.get_json('btcparalelo.com', '/api/price')
-        return {'VEF': Decimal(json['price'])}
+        json = await self.get_json('bylls.com', '/api/price?from_currency=BTC&to_currency=CAD')
+        return {'CAD': Decimal(json['public_price']['to_price'])}
 
 
 class Coinbase(ExchangeBase):
@@ -251,6 +252,24 @@ class Coinbase(ExchangeBase):
         return {ccy: Decimal(rate) for (ccy, rate) in json["data"]["rates"].items()}
 
 
+class CoinCap(ExchangeBase):
+
+    async def get_rates(self, ccy):
+        json = await self.get_json('api.coincap.io', '/v2/rates/bitcoin/')
+        return {'USD': Decimal(json['data']['rateUsd'])}
+
+    def history_ccys(self):
+        return ['USD']
+
+    async def request_history(self, ccy):
+        # Currently 2000 days is the maximum in 1 API call
+        # (and history starts on 2017-03-23)
+        history = await self.get_json('api.coincap.io',
+                                      '/v2/assets/bitcoin/history?interval=d1&limit=2000')
+        return dict([(datetime.utcfromtimestamp(h['time']/1000).strftime('%Y-%m-%d'), h['priceUsd'])
+                     for h in history['data']])
+
+
 class CoinDesk(ExchangeBase):
 
     async def get_currencies(self):
@@ -280,18 +299,23 @@ class CoinDesk(ExchangeBase):
         return json['bpi']
 
 
-class Coinsecure(ExchangeBase):
+class CoinGecko(ExchangeBase):
 
     async def get_rates(self, ccy):
-        json = await self.get_json('api.coinsecure.in', '/v0/noauth/newticker')
-        return {'INR': Decimal(json['lastprice'] / 100.0 )}
+        json = await self.get_json('api.coingecko.com', '/api/v3/exchange_rates')
+        return dict([(ccy.upper(), Decimal(d['value']))
+                     for ccy, d in json['rates'].items()])
 
+    def history_ccys(self):
+        # CoinGecko seems to have historical data for all ccys it supports
+        return CURRENCIES[self.name()]
 
-class Foxbit(ExchangeBase):
+    async def request_history(self, ccy):
+        history = await self.get_json('api.coingecko.com',
+                                      '/api/v3/coins/bitcoin/market_chart?vs_currency=%s&days=max' % ccy)
 
-    async def get_rates(self,ccy):
-        json = await self.get_json('api.bitvalor.com', '/v1/ticker.json')
-        return {'BRL': Decimal(json['ticker_1h']['exchanges']['FOX']['last'])}
+        return dict([(datetime.utcfromtimestamp(h[0]/1000).strftime('%Y-%m-%d'), h[1])
+                     for h in history['prices']])
 
 
 class itBit(ExchangeBase):
@@ -344,23 +368,6 @@ class TheRockTrading(ExchangeBase):
                              '/v1/funds/BTCEUR/ticker')
         return {'EUR': Decimal(json['last'])}
 
-class Unocoin(ExchangeBase):
-
-    async def get_rates(self, ccy):
-        json = await self.get_json('www.unocoin.com', 'trade?buy')
-        return {'INR': Decimal(json)}
-
-
-class WEX(ExchangeBase):
-
-    async def get_rates(self, ccy):
-        json_eur = await self.get_json('wex.nz', '/api/3/ticker/btc_eur')
-        json_rub = await self.get_json('wex.nz', '/api/3/ticker/btc_rur')
-        json_usd = await self.get_json('wex.nz', '/api/3/ticker/btc_usd')
-        return {'EUR': Decimal(json_eur['btc_eur']['last']),
-                'RUB': Decimal(json_rub['btc_rur']['last']),
-                'USD': Decimal(json_usd['btc_usd']['last'])}
-
 
 class Winkdex(ExchangeBase):
 
@@ -394,25 +401,39 @@ def dictinvert(d):
     return inv
 
 def get_exchanges_and_currencies():
+    # load currencies.json from disk
     path = resource_path('currencies.json')
     try:
         with open(path, 'r', encoding='utf-8') as f:
             return json.loads(f.read())
     except:
         pass
+    # or if not present, generate it now.
+    print("cannot find currencies.json. will regenerate it now.")
     d = {}
     is_exchange = lambda obj: (inspect.isclass(obj)
                                and issubclass(obj, ExchangeBase)
                                and obj != ExchangeBase)
     exchanges = dict(inspect.getmembers(sys.modules[__name__], is_exchange))
-    for name, klass in exchanges.items():
-        exchange = klass(None, None)
+
+    async def get_currencies_safe(name, exchange):
         try:
-            d[name] = exchange.get_currencies()
+            d[name] = await exchange.get_currencies()
             print(name, "ok")
         except:
             print(name, "error")
-            continue
+
+    async def query_all_exchanges_for_their_ccys_over_network():
+        async with timeout_after(10):
+            async with TaskGroup() as group:
+                for name, klass in exchanges.items():
+                    exchange = klass(None, None)
+                    await group.spawn(get_currencies_safe(name, exchange))
+    loop = asyncio.get_event_loop()
+    try:
+        loop.run_until_complete(query_all_exchanges_for_their_ccys_over_network())
+    except Exception as e:
+        pass
     with open(path, 'w', encoding='utf-8') as f:
         f.write(json.dumps(d, indent=4, sort_keys=True))
     return d
@@ -436,6 +457,7 @@ def get_exchanges_by_ccy(history=True):
 class FxThread(ThreadJob):
 
     def __init__(self, config: SimpleConfig, network: Network):
+        ThreadJob.__init__(self)
         self.config = config
         self.network = network
         if self.network:
@@ -478,27 +500,28 @@ class FxThread(ThreadJob):
 
     async def run(self):
         while True:
+            # approx. every 2.5 minutes, refresh spot price
             try:
-                await asyncio.wait_for(self._trigger.wait(), 150)
-            except concurrent.futures.TimeoutError:
+                async with timeout_after(150):
+                    await self._trigger.wait()
+                    self._trigger.clear()
+                # we were manually triggered, so get historical rates
+                if self.is_enabled() and self.show_history():
+                    self.exchange.get_historical_rates(self.ccy, self.cache_dir)
+            except TaskTimeout:
                 pass
-            else:
-                self._trigger.clear()
-                if self.is_enabled():
-                    if self.show_history():
-                        self.exchange.get_historical_rates(self.ccy, self.cache_dir)
             if self.is_enabled():
-                self.exchange.update(self.ccy)
+                await self.exchange.update_safe(self.ccy)
 
     def is_enabled(self):
-        return bool(self.config.get('use_exchange_rate'))
+        return bool(self.config.get('use_exchange_rate', DEFAULT_ENABLED))
 
     def set_enabled(self, b):
         self.config.set_key('use_exchange_rate', bool(b))
         self.trigger_update()
 
-    def get_history_config(self):
-        return bool(self.config.get('history_rates'))
+    def get_history_config(self, *, default=False):
+        return bool(self.config.get('history_rates', default))
 
     def set_history_config(self, b):
         self.config.set_key('history_rates', bool(b))
@@ -517,10 +540,10 @@ class FxThread(ThreadJob):
 
     def get_currency(self):
         '''Use when dynamic fetching is needed'''
-        return self.config.get("currency", "EUR")
+        return self.config.get("currency", DEFAULT_CURRENCY)
 
     def config_exchange(self):
-        return self.config.get('use_exchange', 'BitcoinAverage')
+        return self.config.get('use_exchange', DEFAULT_EXCHANGE)
 
     def show_history(self):
         return self.is_enabled() and self.get_history_config() and self.ccy in self.exchange.history_ccys()
@@ -536,8 +559,8 @@ class FxThread(ThreadJob):
             self.network.asyncio_loop.call_soon_threadsafe(self._trigger.set)
 
     def set_exchange(self, name):
-        class_ = globals().get(name, BitcoinAverage)
-        self.print_error("using exchange", name)
+        class_ = globals().get(name) or globals().get(DEFAULT_EXCHANGE)
+        self.logger.info(f"using exchange {name}")
         if self.config_exchange() != name:
             self.config.set_key('use_exchange', name, True)
         self.exchange = class_(self.on_quotes, self.on_history)
@@ -606,3 +629,6 @@ class FxThread(ThreadJob):
         from .util import timestamp_to_datetime
         date = timestamp_to_datetime(timestamp)
         return self.history_rate(date)
+
+
+assert globals().get(DEFAULT_EXCHANGE), f"default exchange {DEFAULT_EXCHANGE} does not exist"
