@@ -39,11 +39,15 @@ except ImportError:
     sys.exit("Error: could not find paymentrequest_pb2.py. Create it with 'protoc --proto_path=electrum/ --python_out=electrum/ electrum/paymentrequest.proto'")
 
 from . import bitcoin, ecc, util, transaction, x509, rsakey
-from .util import print_error, bh2u, bfh, export_meta, import_meta, make_aiohttp_session
+from .util import bh2u, bfh, export_meta, import_meta, make_aiohttp_session
 from .crypto import sha256
 from .bitcoin import TYPE_ADDRESS
 from .transaction import TxOutput
 from .network import Network
+from .logging import get_logger, Logger
+
+
+_logger = get_logger(__name__)
 
 
 REQUEST_HEADERS = {'Accept': 'application/bitcoin-paymentrequest', 'User-Agent': 'Electrum'}
@@ -86,7 +90,7 @@ async def get_payment_request(url: str) -> 'PaymentRequest':
                     else:
                         data = resp_content
                     data_len = len(data) if data is not None else None
-                    print_error('fetched payment request', url, data_len)
+                    _logger.info(f'fetched payment request {url} {data_len}')
         except aiohttp.ClientError as e:
             error = f"Error while contacting payment URL:\n{repr(e)}"
             if isinstance(e, aiohttp.ClientResponseError) and e.status == 400 and resp_content:
@@ -180,7 +184,7 @@ class PaymentRequest:
         try:
             x, ca = verify_cert_chain(cert.certificate)
         except BaseException as e:
-            traceback.print_exc(file=sys.stderr)
+            _logger.exception('')
             self.error = str(e)
             return False
         # get requestor name
@@ -454,9 +458,10 @@ def make_request(config, req):
 
 
 
-class InvoiceStore(object):
+class InvoiceStore(Logger):
 
     def __init__(self, storage):
+        Logger.__init__(self)
         self.storage = storage
         self.invoices = {}
         self.paid = {}
@@ -511,7 +516,7 @@ class InvoiceStore(object):
     def get_status(self, key):
         pr = self.get(key)
         if pr is None:
-            print_error("[InvoiceStore] get_status() can't find pr for", key)
+            self.logger.info(f"get_status() can't find pr for {key}")
             return
         if pr.tx is not None:
             return PR_PAID
@@ -537,4 +542,7 @@ class InvoiceStore(object):
         return self.invoices.values()
 
     def unpaid_invoices(self):
-        return [ self.invoices[k] for k in filter(lambda x: self.get_status(x)!=PR_PAID, self.invoices.keys())]
+        return [self.invoices[k] for k in
+                filter(lambda x: self.get_status(x) not in (PR_PAID, None),
+                       self.invoices.keys())
+                ]
