@@ -1,10 +1,12 @@
 import time
 from struct import pack
 
+from electrum import ecc
 from electrum.i18n import _
-from electrum.util import PrintError, UserCancelled
+from electrum.util import UserCancelled
 from electrum.keystore import bip39_normalize_passphrase
-from electrum.bip32 import serialize_xpub, convert_bip32_path_to_list_of_uint32
+from electrum.bip32 import BIP32Node, convert_bip32_path_to_list_of_uint32
+from electrum.logging import Logger
 
 
 class GuiMixin(object):
@@ -94,7 +96,7 @@ class GuiMixin(object):
         return self.proto.WordAck(word=word)
 
 
-class SafeTClientBase(GuiMixin, PrintError):
+class SafeTClientBase(GuiMixin, Logger):
 
     def __init__(self, handler, plugin, proto):
         assert hasattr(self, 'tx_api')  # ProtocolMixin already constructed?
@@ -105,6 +107,7 @@ class SafeTClientBase(GuiMixin, PrintError):
         self.types = plugin.types
         self.msg = None
         self.creating_wallet = False
+        Logger.__init__(self)
         self.used()
 
     def __str__(self):
@@ -138,7 +141,7 @@ class SafeTClientBase(GuiMixin, PrintError):
     def timeout(self, cutoff):
         '''Time out the client if the last operation was before cutoff.'''
         if self.last_operation < cutoff:
-            self.print_error("timed out")
+            self.logger.info("timed out")
             self.clear_session()
 
     @staticmethod
@@ -156,7 +159,12 @@ class SafeTClientBase(GuiMixin, PrintError):
         address_n = self.expand_path(bip32_path)
         creating = False
         node = self.get_public_node(address_n, creating).node
-        return serialize_xpub(xtype, node.chain_code, node.public_key, node.depth, self.i4b(node.fingerprint), self.i4b(node.child_num))
+        return BIP32Node(xtype=xtype,
+                         eckey=ecc.ECPubkey(node.public_key),
+                         chaincode=node.chain_code,
+                         depth=node.depth,
+                         fingerprint=self.i4b(node.fingerprint),
+                         child_number=self.i4b(node.child_num)).to_xpub()
 
     def toggle_passphrase(self):
         if self.features.passphrase_protection:
@@ -186,13 +194,13 @@ class SafeTClientBase(GuiMixin, PrintError):
     def clear_session(self):
         '''Clear the session to force pin (and passphrase if enabled)
         re-entry.  Does not leak exceptions.'''
-        self.print_error("clear session:", self)
+        self.logger.info(f"clear session: {self}")
         self.prevent_timeouts()
         try:
             super(SafeTClientBase, self).clear_session()
         except BaseException as e:
             # If the device was removed it has the same effect...
-            self.print_error("clear_session: ignoring error", str(e))
+            self.logger.info(f"clear_session: ignoring error {e}")
 
     def get_public_node(self, address_n, creating):
         self.creating_wallet = creating
@@ -200,7 +208,7 @@ class SafeTClientBase(GuiMixin, PrintError):
 
     def close(self):
         '''Called when Our wallet was closed or the device removed.'''
-        self.print_error("closing client")
+        self.logger.info("closing client")
         self.clear_session()
         # Release the device
         self.transport.close()
